@@ -24,6 +24,7 @@ class DataStorageType(FireStoreBase):
         self._ext_storage_fields = {key:StorageField.getInstance(self.ext_storage_fields[key]) for key in self.ext_storage_fields}
         super(DataStorageType, self).__init__(fsClient, **kwargs)
         DataStorageType.DST_STORE[self.storage_name] = self
+        
 
     def base_path(self):
         return DataStorageType.__basePath(self._fsClient)
@@ -52,6 +53,15 @@ class DataStorageType(FireStoreBase):
             docDict['fs_docRef'] = ref
         return DataStorageType(DataStorageType.get_firestore_client(), **docDict)
 
+    @classmethod
+    def getExtendsInstance(cls, dataStorageType, extension):
+        args = {}
+        args['storage_name'] = dataStorageType.storage_name
+        args['isList'] = dataStorageType.isList
+        args['extends'] = extension
+        args['storage_fields'] = dataStorageType.storage_fields
+        return DataStorageType(dataStorageType._fsClient,**args)
+    
     @classmethod
     def get_all_field_containers(cls):
         dst_coll = DataStorageType.get_firestore_client().collection(DataStorageType.basePath())
