@@ -30,7 +30,56 @@ class SalesInventoryAPI(object):
         if logger is None:
             logger = logging.getLogger('SalesInventoryAPI')
         self.logger = logger
+    
+    def _reserveNotFoundErr(self,reserveId):
+        return self._logRaiseReserveError(reserveId,"The look could not be found to load the reserve.")
+    
+    def _logRaiseReserveError(self, reserveId:str, msg:str):
+        msg = f"Problem with Reserve: {reserveId} -- {msg}"
+        return self._logRaiseError(msg)
+    
+    def _logRaiseError(self, errMessage:str):
+        print(f"!!!ERROR!!!ERROR!!!__{errMessage}")
+        self.logger.error(errMessage)
+        return Exception(errMessage)
+    
+    def copyReserveItem(self,reserveId:str,growWeek:str=None) -> dict:
+        '''
+        Take the plant item and make a copy that is specific for this reserve
+        
+        Parameters
+        ----------
+            reserveId : str
+                The reserve id number to create a custom item for
+            
+        '''
+        resp = ItemReserve.CreateCustomItem(reserveId,growWeek=growWeek)
+        if resp is None:
+            raise self._reserveNotFoundErr(reserveId)
+        
+        summ = ReserveSummary.getReserveSummary(resp['reserve'].finish_week)
+        summUpdateResp = summ.update_reserve(resp['reserve'])
+        return {'reserve_update':summUpdateResp, 'new_item': resp['item']}
 
+    def refreshReserve(self,reserveId:str, growWeek:str = None) -> dict:
+        '''
+        Refresh a single reserve and update the summary
+        
+        Parameters
+        ----------
+            reserveId : str
+                The reserve id for the reserve we want to refresh
+            
+        '''
+        
+        ir = ItemReserve.getItemReserveInstance(reserveId, growWeek)
+        if ir is None:
+            raise self._reserveNotFoundErr(reserveId)
+        
+        summ = ReserveSummary.getReserveSummary(ir.finish_week)
+        summ.update_reserve(ir)
+        return ir.get_dict()
+    
     def createReserveAPI(self, reserveDate, customer_id : str, location_id : str, product_id : str, reserved : int, invLoc=None) -> list:
         '''
         create a reserve
@@ -56,7 +105,14 @@ class SalesInventoryAPI(object):
         summ = ReserveSummary.getReserveSummary(gw.id)
         return summ.add_reserve(ir)
 
-    def updateReserveAPI(self, reserve_id : str, customer_id : str, location_id : str, product_id : str, reserved : int, reserveDate : datetime, invLoc=None) -> list:
+    def updateReserveAPI(self, reserve_id : str, 
+                         customer_id : str, 
+                         location_id : str, 
+                         product_id : str, 
+                         reserved : int, 
+                         reserveDate : datetime, 
+                         invLoc:str=None,
+                         origGrowWeek:str=None) -> list:
         '''
         Update the reserve given the reserve id
 
@@ -79,12 +135,14 @@ class SalesInventoryAPI(object):
             
         '''
         gw = GrowWeek.GetGrowWeekByDate(reserveDate)
-        ir = ItemReserve.getItemReserveInstance(reserve_id)
+        ir = ItemReserve.getItemReserveInstance(reserve_id,origGrowWeek)
+        if ir is None:
+            raise self._reserveNotFoundErr(reserve_id)
         ir = ir.update_reserve(customer_id, location_id, product_id, reserved, gw, reserveDate, invLoc)
         summ = ReserveSummary.getReserveSummary(gw.id)
         return summ.update_reserve(ir)
 
-    def deleteReserveAPI(self, reserve_id : str) -> dict:
+    def deleteReserveAPI(self, reserve_id : str, growWeek : str = None) -> dict:
         '''
         Delete the reserve given the reserve id
 
@@ -93,21 +151,15 @@ class SalesInventoryAPI(object):
             reserve_id : str
                 The string key that is unique identifier for a reserve
         '''
-        ir = ItemReserve.getItemReserveInstance(reserve_id)
-        if ir is not None and ir.exists:
-            gw = GrowWeek.getGrowWeekInstance(ir.grow_period)
-            summ = ReserveSummary.getReserveSummary(gw.id)
-            summ.delete_reserve(ir)
-            ItemReserve.hardDelete(ir.id)
-            return {"status": "success"}
-        else:
-            path = DataNumberLookup.get_path_for_dnl(reserve_id)
-            docRef = ir._fsClient.document(path)
-            gw_id = docRef.parent.parent.id
-            gw = GrowWeek.getGrowWeekInstance(gw_id)
-            summ = ReserveSummary.getReserveSummary(gw.id)
-            summ._delete_reserve(reserve_id)
-            return {"status": "success"}
+        ir = ItemReserve.getItemReserveInstance(reserve_id, growWeek)
+        if ir is None:
+            raise self._reserveNotFoundErr(reserve_id)
+
+        gw = GrowWeek.getGrowWeekInstance(ir.grow_period)
+        summ = ReserveSummary.getReserveSummary(gw.id)
+        summ.delete_reserve(ir)
+        ItemReserve.hardDelete(ir.id)
+        return {"status": "success"}
 
     def getAllReservesAPI(self, reserveDate : datetime) -> list:
         '''
@@ -145,6 +197,8 @@ class SalesInventoryAPI(object):
                 The string key to this reserve
         '''
         ir = ItemReserve.getItemReserveInstance(reserve_id)
+        if ir is None:
+            raise self._reserveNotFoundErr(reserve_id)
         return ir.get_dict()
 
     def process_productionview_item_update(self, item_id : str, status : str) -> dict:
