@@ -21,15 +21,23 @@ class StorageBlob(FireStoreBase):
 
     COLLECTION_NAME = 'application_data'
     ext_fields = ['key_fields','data_number_lookup','data_type']
-    CPI = 'cust_plant_item.*.{"Customer_Name": customer_name, "Customer_Id": customer_id, "Location_Name": location_name, "Location_Id": location_id, "Type":`Customer`,"Product_Name":Product_Name,"Item_Num":Cust_Item_Num,"plant_image":plant_image[0].image_url}'
-    PI = 'plant_item.*.{"Customer_Name": `master`, "Customer_Id": ``, "Location_Name": `master`, "Location_Id":   ``, "Type":`Master`,"Product_Name":Product_Name,"Item_Num":Item_Num,"plant_image":plant_image[0].image_url}'
+    CPI_FLDS = '{"Customer_Name": customer_name, "Customer_Id": customer_id, "Location_Name": location_name, "Location_Id": location_id, "Type":`Customer`,"Product_Name":Product_Name,"Item_Num":Cust_Item_Num,"plant_image":plant_image[0].image_url}'
+    CPI = 'cust_plant_item.*.'+CPI_FLDS
+    PI_FLDS = '{"Customer_Name": `master`, "Customer_Id": ``, "Location_Name": `master`, "Location_Id":   ``, "Type":`Master`,"Product_Name":Product_Name,"Item_Num":Item_Num,"plant_image":plant_image[0].image_url}'
+    PI = 'plant_item.*.'+PI_FLDS
+
+    TYPE_FIELDS = {}
 
 
     def __init__(self, fsClient, **kwargs):
         self.key_fields = kwargs.get('key_fields','') #ndb.StringProperty(required=True)
         self.data_number_lookup = kwargs.get('data_number_lookup','') #ndb.StringProperty()
         #self.data = ndb.JsonProperty()
-        self.data_type = kwargs.get('data_type','') #ndb.StringProperty(required=True)  # Refers to the DataStorageType.storage_name
+        dType = ''
+        if kwargs.get('fs_docRef',None) is not None: # Updated to pull the collection name... this should be the data type for a storage blog, this defaults if it isn't in the field
+            if kwargs['fs_docRef'].parent:
+                dType = kwargs['fs_docRef'].parent.id
+        self.data_type = kwargs.get('data_type',dType) #ndb.StringProperty(required=True)  # Refers to the DataStorageType.storage_name
         self._non_data_fields = FireStoreBase.base_fields+['key_fields','data_number_lookup','data_type','fs_docRef','fs_docSnap']
         self._parent = None
         # This part of the code is designed to help load data passed in...
@@ -38,11 +46,12 @@ class StorageBlob(FireStoreBase):
         keys = kwargs.keys()
         xtra_fields = [x for x in keys if x not in self._non_data_fields]
         extraDict = {}
-        self._sb_fields = []
+        self._sb_fields = self._load_dst_fields()
 
         for k in xtra_fields:
             extraDict[k] = kwargs[k]
-            self._sb_fields.append(k)
+            if k not in self._sb_fields:
+                self._sb_fields.append(k)
 
         self.update_data(extraDict)
         super(StorageBlob, self).__init__(fsClient, **kwargs)
@@ -72,11 +81,21 @@ class StorageBlob(FireStoreBase):
     @classmethod
     def getInstanceByPath(cls,path):
         return StorageBlob.getInstance(StorageBlob.get_firestore_client().document(path))
+    
+    @classmethod
+    def getInstanceByDNL(cls,dnl):
+        path = DataNumberLookup.get_path_for_dnl(dnl)
+        return StorageBlob.getInstanceByPath(path)
 
     @classmethod
     def get_by_dnl(cls,dataNumberLookup):
         return DataNumberLookup.get_obj_by_dnl(dataNumberLookup,StorageBlob)
 
+    def _load_dst_fields(self,dst=None):
+        if dst is None:
+            dst = DataStorageType.get_dataStorageType(self.data_type)
+        return list(dst._ext_storage_fields.keys())
+
     def get_id_ancestry(self,ndbKey, idString=""):
         '''
         Won't implement this... but will create get "get_parent_path"
@@ -102,6 +121,27 @@ class StorageBlob(FireStoreBase):
                 self._parent = StorageBlob.getInstance(parColl.parent)
 
         return self._parent
+    
+    def get_custom_dict(self,lookupDict):
+        '''
+        If you need a custom dictionary, send in this request
+
+        Parameters
+        ----------
+            lookupDict : list(dict)
+                key_name : str
+                    The name of the key for the field
+                key_value : str
+                    The actual field name
+        '''
+        respDict = {}
+        for lookupEntry in lookupDict:
+            if 'key_name' not in lookupEntry.keys():
+                raise Exception("Invalid request was made to get_custom_dict in StorageBlob")
+            fieldName = lookupEntry['key_name']
+            fieldValue = getattr(self,lookupEntry.get('key_value',fieldName),None)
+            respDict[fieldName] = fieldValue
+        return respDict
 
     def get_sb_data(self):
         '''
@@ -142,14 +182,19 @@ class StorageBlob(FireStoreBase):
         return StorageBlob.getInstance(StorageBlob.get_firestore_client().document(path))
 
     @classmethod
-    def get_SB_from_Collection(cls,parent_id,model_name):
+    def get_SB_from_Collection(cls,parent_id,model_name,doFilter=False):
         docRefs = []
         if parent_id is not None and parent_id != '':
-            docRefs = DataNumberLookup.get_doc_children(parent_id,model_name)
+            docRefs = DataNumberLookup.get_doc_children(parent_id,model_name,doFilter)
         else:
             sbPath = StorageBlob.basePath()+'/'+model_name
             collection = StorageBlob.get_firestore_client().collection(sbPath)
-            docRefs = collection.list_documents()
+            if not doFilter:
+                docRefs = collection.list_documents()
+            else:
+                q = collection.where("soft_delete","==",doFilter)
+                snaps = q.stream()
+                docRefs = [x for x in snaps]
         
         return [StorageBlob.getInstance(docRef) for docRef in docRefs]
 
@@ -171,34 +216,63 @@ class StorageBlob(FireStoreBase):
     '''
     @classmethod
     def get_all_plant_items(cls,use_cache=True):
-        # TODO implement a call to cache..
-        #resp = memcache.get(ALL_PLANT_ITEMS)
-        #resp = QuickStorage.getValue(ALL_PLANT_ITEMS)
-        #if resp and use_cache:
-            #return StorageBlob.render_array_response(resp)
-        #    return resp
+        resp = cls._get_api_qs_data()
+        if resp and use_cache:
+            return resp
 
         logging.info("Cached results not found for plant info... pulling data.. (USE CACHE: "+str(use_cache)+")")
-        resp = {'plant_item':{},'cust_plant_item':{}}
+        arr_resp = []
 
-        t = 'plant_item'
-        #pi = StorageBlob.query(StorageBlob.data_type == t)
-        pi = StorageBlob.get_firestore_client().collection_group(t)
-        for sbSnap in pi.stream():
-            sb = StorageBlob.getInstance(sbSnap)
-            resp[t][sb.data_number_lookup] = sb.get_plant_item_data()
+        mstr_path = 'application_data/Color_Orchids/Customer_Tracking/StorageBlob/admin_items/admin_items-100/master_items'
+        pi = StorageBlob.get_firestore_client().collection(mstr_path)
+        docs = pi.list_documents()
+        docRefs = [doc for doc in docs]
+        print("Found: "+str(len(docRefs))+" master items..." )
+        for docRef in docRefs:
+            arr_resp.append(StorageBlob.getInstance(docRef)._render_api_summary(cpi=False))
 
-        t = 'cust_plant_item'
-        pi = StorageBlob.get_firestore_client().collection_group(t)
-        for sbSnap in pi.stream():
-            sb = StorageBlob.getInstance(sbSnap)
-            resp[t][sb.data_number_lookup] = sb.get_plant_item_data()
+        pi = StorageBlob.get_firestore_client().collection_group('items')
+        snps = [snap for snap in pi.stream()]
+        print("Found: "+str(len(snps))+" customer plant items..." )
+        for sbSnap in snps:
+            arr_resp.append(StorageBlob.getInstance(sbSnap)._render_api_summary(cpi=True))
 
-        #memcache.set(ALL_PLANT_ITEMS, resp, DEFAULT_CACHE_TIME * 24)
-        #QuickStorage.setValue(ALL_PLANT_ITEMS,resp)
-        # TODO implement some kind of caching render_resp = StorageBlob.set_memcache_pi_values(resp)
+        cls._set_api_qs_data(arr_resp)
 
-        return StorageBlob.__render_array_response(resp) 
+        return arr_resp
+    
+    @classmethod
+    def _get_api_qs_data(cls):
+        allKey = 'ALL_PLANT_ITEMS'
+        qsEntries = QuickStorage.getValue('ALL_PLANT_ITEMS')
+        if qsEntries is None:
+            return None
+        resp = []
+        for qsEntry in qsEntries:
+            entryValue = QuickStorage.getValue(qsEntry)
+            if entryValue is not None:
+                resp = resp + entryValue
+        return resp
+    
+    @classmethod
+    def _set_api_qs_data(cls,responses):
+        allKey = 'ALL_PLANT_ITEMS'
+        count = 0
+        qsArr = []
+        qsKeys = []
+        qsKey = allKey+str(count).zfill(5)
+        for response in responses:
+            if len(qsArr) >= 300:
+                qsKeys.append(qsKey)
+                QuickStorage.setValue(qsKey,qsArr,expireMins=1380)
+                qsKey = allKey+str(count).zfill(5)
+                qsArr = []
+            qsArr.append(response)
+            count += 1
+        if len(qsArr) > 0:
+            QuickStorage.setValue(qsKey,qsArr,expireMins=1380)
+            qsKeys.append(qsKey)
+        QuickStorage.setValue(allKey,qsKeys,expireMins=1380)
 
     @classmethod
     def __render_array_response(cls,dict_resp):
@@ -213,28 +287,28 @@ class StorageBlob(FireStoreBase):
 
     @classmethod
     def get_plant_item(cls, item_number):
-        dnl = StorageBlob.get_firestore_client().document(DataNumberLookup.basePath()+'/'+item_number)
-        doc = dnl.get()
-        if doc.exists:
-            sbRef = StorageBlob.get_firestore_client().document(doc.get('path'))
-            sb = StorageBlob.getInstance(sbRef)
-        if not sb:
+        #dnl = StorageBlob.get_firestore_client().document(DataNumberLookup.basePath()+'/'+item_number)
+        sb = StorageBlob.get_by_dnl(item_number)
+        if not sb.exists:
             return None
 
         return sb.get_plant_item_data()
+    
+    def _render_api_summary(self,cpi=True):
+        apiData = self.get_plant_item_data()
+        r = None
+        if cpi:
+            r = jmespath.search(StorageBlob.CPI_FLDS,apiData)
+        else:
+            r = jmespath.search(StorageBlob.PI_FLDS,apiData)
+        return r
+
 
     def get_plant_item_data(self):
         pi_data = self.get_sb_data()
 
         pi_images = pi_data.get('plant_image',[])
-        image_details = []
-        if pi_images:
-            for image in pi_images:
-                if image:
-                    image_detail = Image.getImageInfo(image)
-                    image_details.append(image_detail)
-
-        pi_data['plant_image'] = image_details
+        pi_data['plant_image'] = [] if len(pi_images) == 0 else pi_images
         par = self.get_parent()
         if par and par.data_type == 'location':
             self.get_plant_ancestor_info(pi_data)
@@ -323,6 +397,11 @@ class StorageBlob(FireStoreBase):
                     results[option_type].remove(item)
                 else:
                     item.update(StorageBlob.__convert_to_lookup(option_type,StorageBlob(None,**option_update)))
+            else:
+                if not doDelete:
+                    entry = StorageBlob.__convert_to_lookup(option_type,StorageBlob(None,**option_update))
+                    results[option_type].append(entry)
+                    
             QuickStorage.setValue(qsKey,results,1080,True)
 
     @classmethod
@@ -506,10 +585,6 @@ class StorageBlob(FireStoreBase):
         '''
         All option data will be at the base level... so no need to do a collection group
         '''
-        jmKey = storeName+jmPath
-        qHits = QuickStorage.getValue(jmKey)
-        if qHits is not None:
-            return qHits
 
         where_stmts = StorageBlob.parse_jmespath(jmPath)
         sbPath = StorageBlob.basePath()+'/'+storeName
@@ -537,8 +612,6 @@ class StorageBlob(FireStoreBase):
             option['value'] = str(data['id']) +"|"+ value if includeId else value
             returnOptions.append(option)
 
-        QuickStorage.setValue(jmKey,returnOptions,360)
-
         return returnOptions
 
     @classmethod
@@ -574,7 +647,12 @@ class StorageBlob(FireStoreBase):
 
     @classmethod
     def get_opdata_options(cls,option_container,option_fld):
-        _, opFld = option_fld.split(".")
+        opFld = None
+        if len(option_fld.split(".")) > 1:
+            _, opFld = option_fld.split(".")
+        else:
+            opFld = option_fld
+            
         option_field = replace_op_container(option_container)
         qHits = QuickStorage.getValue(option_field)
         if qHits is not None:
@@ -612,7 +690,7 @@ class StorageBlob(FireStoreBase):
         if errs == '':
             resp = super(StorageBlob,self).delete_resp()
             if not resp['didDelete']:
-                logger.warn(resp['msg'])
+                logging.warn(resp['msg'])
         else:
             resp['didDelete'] = False
             resp['status'] = 'failed'
@@ -641,6 +719,9 @@ class StorageBlob(FireStoreBase):
     def validate_data(self, dst=None, fill_missing = False):
         if not dst:
             dst = DataStorageType.get_dataStorageType(self.data_type)
+            if dst is not None:
+                if dst.storage_name == 'Copies':
+                    dst = DataStorageType.getExtendsInstance(dst,self.orig_data_type)
 
         if not dst:
             raise Exception("The data storage name of: "+self.data_type+", does not exist")
@@ -672,8 +753,9 @@ class StorageBlob(FireStoreBase):
     '''
     Shouldn't have more than 1 auto_inc fields in a storagetype
     '''
-    def create_steps(self):
-        dst = DataStorageType.get_dataStorageType(self.data_type)
+    def create_steps(self,dst=None):
+        if dst is None:
+            dst = DataStorageType.get_dataStorageType(self.data_type)
 
         self.validate_data(dst=dst,fill_missing=True)
 
@@ -688,6 +770,19 @@ class StorageBlob(FireStoreBase):
 
         return dnVal
 
+    def update_sbobj(self,updateDict):
+        '''
+        Given a dictionary with updated data.. set this storage blob up and validate and save
+
+        Parameters
+        ----------
+            updateDict : dict
+                The key/value pair dictionary with fields to be updated
+        '''
+        self.update_data(updateDict)
+        self.validate_data()
+        self.update_ndb()
+
 
     @classmethod
     def create_blob_parent(cls, data_type, blob_data, parentId, collection=None):
@@ -695,7 +790,36 @@ class StorageBlob(FireStoreBase):
             raise Exception("This method is intended to create a storageblob under a parent... use a different method if you are creating from scratch")
 
         return StorageBlob.create_blob_parent_fromPath(data_type,blob_data,DataNumberLookup.get_path_for_dnl(parentId),collection)
+    
+    @classmethod
+    def createItemCopy(cls, copyDnl, dnlExt=None):
+        copySb = cls.get_by_dnl(copyDnl)
+        if copySb is not None:
+            return copySb._createItemCopy(dnlExt)
+        return None
+    
+    def _createItemCopy(self,dnlExt=None):
+        if dnlExt is None:
+            dnlExt = datetime.now().isoformat('-').replace('-','').replace(':','').replace('.','')
+            
+        copyDataType = self.data_type
+        newDnl = f'Copies-{self.data_number_lookup}__{dnlExt}'
+        
+        dst = DataStorageType.get_dataStorageType('Copies') # officially added on 2/5/2022
+        dstExt = DataStorageType.getExtendsInstance(dst,copyDataType)
+        
+        blob_data = {"orig_path":self.path,
+                     "orig_dnl": self.data_number_lookup,
+                     "orig_data_type": copyDataType,
+                     "isCopy":True}
+        
+        sb_fields = list(dstExt._ext_storage_fields.keys())
 
+        for field in sb_fields:
+            if field not in list(blob_data.keys()): # Don't want to overwrite
+                blob_data[field] = getattr(self,field,"")
+        
+        return StorageBlob._create_blob_parent_fromPath(dstExt,blob_data,inDnl=newDnl)
 
     @classmethod
     def get_next_dnl(cls,data_type):
@@ -712,9 +836,18 @@ class StorageBlob(FireStoreBase):
         if not dst:
             raise Exception("The data storage name of: "+data_type+", does not exist")
         
+        return cls._create_blob_parent_fromPath(cls,dst, blob_data, par_path, collection, None)
+    
+    @classmethod
+    def _create_blob_parent_fromPath(cls, dst, blob_data, par_path = None, collection = None, inDnl = None):
+        data_type = dst.storage_name
+        
         clt = StorageBlob.get_client()
         docRef = None
-        dnl = StorageBlob.get_next_dnl(data_type)
+        dnl = inDnl
+        
+        if dnl is None:
+            dnl = StorageBlob.get_next_dnl(data_type)
         if par_path:
             par_doc = clt.fsClient.document(par_path)
             collName = data_type if collection is None else collection
@@ -729,7 +862,7 @@ class StorageBlob(FireStoreBase):
         blob_data['data_type'] = data_type
 
         sb = StorageBlob(clt.fsClient,**blob_data)
-        sb.create_steps()
+        sb.create_steps(dst)
 
         sb.update_ndb(True)
 
@@ -753,5 +886,5 @@ class StorageBlob(FireStoreBase):
             blob.update_data(blob_data)
             blob.validate_data()
             blob.update_ndb()
-            return blob.get_sb_data
+            return blob.get_sb_data()
         return {'data_type':'','id':'','data':{}}
\ No newline at end of file
