@@ -232,6 +232,15 @@ class FSDocument(object):
         else:
             return self.ref.set(field_updates)
     
+    def setSoftDelete(self, softDelete=True):
+        self.ref.update({"soft_delete":softDelete})
+        
+    def makeInactive(self):
+        self.setSoftDelete()
+    
+    def activate(self):
+        self.setSoftDelete(False)
+    
     def to_json(self):
         return {'object':'FSDocument','path':self.path}
     
